#include "decoder.h"
#include <asterisk/lock.h>
static int resample(void *handle, float * src, int srclen, float *dst, double factor, int last)
{
	int expectedlen = (int)(srclen * factor);
	int dstlen = expectedlen + 1000;
	int o, srcused, srcpos, out;

	o = out = srcused = srcpos = 0;
	
	for(;;) {
		int srcBlock = MIN(srclen-srcpos, srclen);
		//int lastFlag = (last && (srcBlock == srclen-srcpos));
		//fprintf(stderr, "resampling %d/%d (%d)\n",  srcpos, srclen,  MIN(dstlen-out, dstlen));
		o = resample_process(handle, factor, &src[srcpos], srcBlock, 0, &srcused, &dst[out], MIN(dstlen-out, dstlen));
		srcpos += srcused;
		if (o >= 0)
			out += o;
		if (o < 0 || (o == 0 && srcpos == srclen))
			break;
	}
        return out;
}

static int init_handler(void ** fh, int type, FILE *in, int *samplerate)
{
	if (type == TYPE_MP3) {
#ifdef MPGLIB
		return init_mp3(fh, in, samplerate);
#else
		//fprintf(stderr, "ERROR: mp3 not supported\n");
#endif
	}
	if (type == TYPE_VORBIS) {
#ifdef OGGVORBIS
		return init_vorbis(fh, in, samplerate);
#else
		//fprintf(stderr, "ERROR: ogg not supported\n");
#endif
	}
	return(0);
}

extern struct decoder * initialize_decoder(int buflen)
{
	struct decoder *decoder = (struct decoder *)malloc(sizeof(struct decoder));
	FILE *in;
	unsigned int seed;
	in = fopen("/dev/urandom", "r");
	fread(&seed, sizeof(int), 1, in);
	fclose(in);
	srand(seed);
	decoder->buffs.output = (short *)malloc(buflen * sizeof(short));
	decoder->buffs.slin8khz = (float *)malloc(buflen * sizeof(float));
	decoder->buffs.input = (float *)malloc(buflen * sizeof(float));
	decoder->random = decoder->buffs.outputlen = decoder->infile = 0;
	decoder->curfile = -1;
	decoder->buflen = buflen;
	return decoder;
}

extern int destroy_decoder(struct decoder * moh)
{
	ast_verbose(VERBOSE_PREFIX_3 "moh closing file: %s\n", moh->files[moh->curfile]);
	if (moh->infile) {
		//fprintf(stderr, "moh closing file: %s\n", moh->files[moh->curfile]);
		close_handler(&moh->buffs.data, moh->curtype);
		//fprintf(stderr, "moh closed handler\n");
		resample_close(moh->buffs.resampler);
		ast_verbose(VERBOSE_PREFIX_3 "moh closed resampler\n");
		//fprintf(stderr, "moh closed resampler\n");
	}
	moh->infile = 0;
	moh->buffs.outputlen = 0;
	free(moh->buffs.output);
	free(moh->buffs.slin8khz);
	free(moh->buffs.input);
	moh->curfile = -1;
	free(moh);
	return 1;
}

extern int read_data(struct decoder * moh, short sbuf[], int size)
{
	int samplerate, len, ret, buflen;
	char filename[256];
	buflen = moh->buflen;
	samplerate = len = ret = 0;
	if (moh->infile==0) {
		ast_verbose(VERBOSE_PREFIX_3 "moh opening new file\n");
		//fprintf(stderr, "moh opening new file\n");
		if (moh->random) {
			moh->curfile = (int)((float)moh->numfiles*rand()/RAND_MAX);
		} else {
			moh->curfile++;
			if (moh->curfile >= moh->numfiles) {
				moh->curfile = 0;
			}
		}
		ast_verbose(VERBOSE_PREFIX_3 "moh selected file %s", moh->files[moh->curfile]);
		//fprintf(stderr, "moh selected file %s", moh->files[moh->curfile]);
		sprintf(filename, "%s/%s", moh->dir, moh->files[moh->curfile]);
		ast_verbose(VERBOSE_PREFIX_3 "%s\n", filename);
		//fprintf(stderr, "%s\n", filename);
		if ((moh->input = fopen(filename, "r"))) {
			ast_verbose(VERBOSE_PREFIX_3 "moh finding type\n");
			//fprintf(stderr, "moh finding type\n");
			moh->curtype = find_type(moh->files[moh->curfile]);
			ast_verbose(VERBOSE_PREFIX_3 "moh initializing handler\n");
			//fprintf(stderr, "moh initializing handler\n");
			if (!init_handler(&moh->buffs.data, moh->curtype, moh->input, &samplerate))
				return 0;
			moh->curfactor = 8000.0 / samplerate;
			moh->buffs.resampler = resample_open(0, moh->curfactor, moh->curfactor);
			ast_verbose(VERBOSE_PREFIX_3 "moh opened file %d: %s type %d rate %d\n", moh->curfile, filename, moh->curtype, samplerate);
			//fprintf(stderr, VERBOSE_PREFIX_3 "moh opened file %d: %s type %d rate %d\n", moh->curfile, filename, moh->curtype, samplerate);
			moh->infile = 1;
		} else {
			ast_verbose(VERBOSE_PREFIX_3 "moh unable to open file %s: %s\n", filename, strerror(errno));
			//fprintf(stderr, "moh unable to open file %s: %s\n", filename, strerror(errno));
		}
	}
	while (ret < size && moh->infile) {
		if (moh->buffs.outputlen == 0) {
#ifdef MPGLIB
			if (moh->curtype == TYPE_MP3) {
				len = read_mp3(moh->buffs.data, moh->buffs.input, buflen);
			}
#endif
#ifdef OGGVORBIS
			if (moh->curtype == TYPE_VORBIS) {
				len = read_vorbis(moh->buffs.data, moh->buffs.input, buflen);
			}
#endif
			if (len > 0) {
				len = resample(moh->buffs.resampler, moh->buffs.input, len, moh->buffs.slin8khz, moh->curfactor, 0);
				moh->buffs.outputlen = ftos(moh->buffs.slin8khz, moh->buffs.output, len);
			} else {
				len = resample(moh->buffs.resampler, moh->buffs.input, len, moh->buffs.slin8khz, moh->curfactor, 1);
				moh->buffs.outputlen = ftos(moh->buffs.slin8khz, moh->buffs.output, len);
				ast_verbose(VERBOSE_PREFIX_3 "moh closing file: %s\n", moh->files[moh->curfile]);
				//fprintf(stderr, "moh closing file: %s\n", moh->files[moh->curfile]);
				moh->infile = 0;
				close_handler(&moh->buffs.data, moh->curtype);
				ast_verbose(VERBOSE_PREFIX_3 "moh closed handler\n");
				//fprintf(stderr, "moh closed handler\n");
				if (moh->buffs.resampler) {
					resample_close(moh->buffs.resampler);
					ast_verbose(VERBOSE_PREFIX_3 "moh closed resampler\n");
					//fprintf(stderr, "moh closed resampler\n");
				}
				return 0;
			}
		}
		if(moh->buffs.outputlen > 0) {
			if (moh->buffs.outputlen+ret > size) {
				memcpy(sbuf+ret, moh->buffs.output, (size-ret)*2);
				moh->buffs.outputlen -= size-ret;
				memmove(moh->buffs.output, moh->buffs.output + size - ret, moh->buffs.outputlen*2);
				ret += size-ret;
			} else {
				memcpy(sbuf+ret, moh->buffs.output, moh->buffs.outputlen*2);
				ret += moh->buffs.outputlen;
				moh->buffs.outputlen = 0;
			}
		}
	}
	return ret;
}

static int close_handler(void ** fh, int type)
{
#ifdef MPGLIB
	if (type == TYPE_MP3)
		return close_mp3(fh);
#endif
#ifdef OGGVORBIS
	if (type == TYPE_VORBIS)
		return close_vorbis(fh);
#endif
	return(0);
}

extern int find_type(char *filename)
{
    	if (strlen(filename) > 3) {
		if (!strcasecmp(filename + strlen(filename) - 4, ".mp3")) {
#ifdef MPGLIB
			return TYPE_MP3;
#else
			ast_log(LOG_WARNING, "moh file type %d (mp3) not supported : %s\n", TYPE_MP3, filename);
			return TYPE_UNKNOWN;
#endif
		}
		if (!strcasecmp(filename + strlen(filename) - 4, ".ogg")) {
#ifdef OGGVORBIS
			return TYPE_VORBIS;
#else
			ast_log(LOG_WARNING, "moh file type %d (ogg) not supported : %s\n", TYPE_VORBIS, filename);
			return TYPE_UNKNOWN;
#endif
		}
		ast_log(LOG_WARNING, "moh file type not supported : %s\n", filename);
		return TYPE_UNKNOWN;
	}
	return TYPE_UNKNOWN;
}
