#include "mp3.h"
#include "../mpglib/mpg123.h"
#include "../mpglib/mpglib.h"

#define INBUFLEN 8192
#define OUTBUFLEN 16384

struct id3tag {
	char header[10];
	char *data;
	int size;
	char flags;
	int hasfooter;
	int curpos;
};

typedef struct mp3fhstruct {
        char inbuf[INBUFLEN];
        int inbuflen;
        char outbuf[OUTBUFLEN];
        int outbuflen;
        int mpgret;
        int samplerate;
        int channels;
        FILE *in;
        struct mpstr mp;
	struct id3tag id3;
} mp3fh;

static int init_mp3fh(mp3fh ** fh)
{
	*fh = (mp3fh *)malloc(sizeof(mp3fh));
	(*fh)->outbuflen = (*fh)->inbuflen = 0;
	InitMP3(&(*fh)->mp);
	return 1;
}

int getid3v2(mp3fh *vh);

extern int init_mp3(void ** fh, FILE *in, int *samplerate)
{
	mp3fh *vh;
	int seek;
	init_mp3fh((mp3fh **)fh);
	vh = *fh;
	vh->in = in;
	seek = vh->id3.curpos = 0;


	vh->inbuflen = fread(vh->inbuf, 1, INBUFLEN, vh->in);
	if ((seek = getid3v2(vh)) > 0) {
		if (vh->inbuflen>seek) {
			vh->inbuflen -= seek;
			memmove(vh->inbuf, vh->inbuf+seek, vh->inbuflen);
		} else if (vh->inbuflen==seek) {
			vh->inbuflen = 0;
		} else {
			vh->inbuflen = 0;
			fseek(vh->in, seek, SEEK_SET);
		}
	}
/*
	int i;
	unsigned char c;
	while (1) {
		if (vh->inbuflen==0)
			vh->inbuflen = fread(vh->inbuf, 1, INBUFLEN, vh->in);
		for(i=0;i<vh->inbuflen;i++) {
			if (   									// Try to find an MP3 header, to skip over any WAV header etc
				( (unsigned char)vh->inbuf[i]==0xff ) 				// First 8 bits must be 0xFF
				&& (c=(unsigned char)vh->inbuf[i+1]) && ((c&0xE0)==0xE0)	// Next 3 bits must be all be 1s
				&& (c=(unsigned char)vh->inbuf[i+1]) && (c&0x06)		// Bits 14-15 must not be 00
				&& (c=(unsigned char)vh->inbuf[i+2]) && ((c&0xF0)!=0xF0)	// Bits 17-20 must not all be 1s
				&& (c=(unsigned char)vh->inbuf[i+2]) && ((c&0x0C)!=0x0C)	// Bits 21-22 must not all be 1s
			) {
				// Found a header...
				if (i==0)
					break;
				vh->inbuflen -= i;
				memmove(vh->inbuf, vh->inbuf+i, vh->inbuflen);
				break;
			}
		}
		if(i<vh->inbuflen || vh->inbuflen<INBUFLEN)
			break;
	}
*/
	do {
		vh->mpgret = decodeMP3(&vh->mp, NULL, 0, vh->outbuf, OUTBUFLEN, &vh->outbuflen);
		if (vh->mpgret == MP3_NEED_MORE) {
			if (vh->inbuflen==0)
				vh->inbuflen = fread(vh->inbuf, 1, INBUFLEN, vh->in);
			if (!vh->inbuflen)
				break;
			vh->mpgret = decodeMP3(&vh->mp, vh->inbuf, vh->inbuflen, vh->outbuf, OUTBUFLEN, &vh->outbuflen);
			vh->inbuflen = 0;
		} 
		if (vh->mpgret == MP3_ERR)
			return 0;
	} while (vh->outbuflen == 0);
	vh->samplerate = (int)freqs[vh->mp.fr.sampling_frequency];
	vh->channels = vh->mp.fr.stereo;
	vh->mp.fr.single = 1;
	*samplerate = vh->samplerate;
	return 1;
}

extern int read_mp3(void *fh, float * buf, int buflen)
{
	mp3fh *vh = fh;
	int written, wlen;

	written = wlen = vh->inbuflen = 0;

	if (vh->outbuflen == 0) {
		do {
			if (vh->mpgret == MP3_OK) {
				vh->mpgret = decodeMP3(&vh->mp, NULL, 0, vh->outbuf, OUTBUFLEN, &vh->outbuflen);
			} else {
				vh->inbuflen = fread(vh->inbuf, 1, INBUFLEN, vh->in);
				if (vh->inbuflen <= 0 && (feof(vh->in) || ferror(vh->in))) {
					return 0;
				} else {
					vh->mpgret = decodeMP3(&vh->mp, vh->inbuf, vh->inbuflen, vh->outbuf, OUTBUFLEN, &vh->outbuflen);
					vh->inbuflen = 0;
					if (vh->mpgret == MP3_ERR)
						return 0;
				}
			}
		} while (vh->outbuflen == 0);
	}
	if (vh->outbuflen == -1) {
		fprintf(stderr, "decoding error!\n");
		vh->outbuflen = 0;
	}

	wlen = MIN(vh->outbuflen,buflen);
	written = ctof(vh->outbuf, buf, wlen);
	memmove(vh->outbuf, vh->outbuf+wlen, (vh->outbuflen-wlen)*sizeof(char));
	vh->outbuflen -= wlen;
	return written;
}

extern int close_mp3(void **fh)
{
	mp3fh *vh = *fh;
	ExitMP3(&vh->mp);
	fclose(vh->in);
	free(*fh);
	return 1;
}

int getid3v2(mp3fh *vh)
{
	if ((vh->inbuflen + vh->id3.curpos) < 10) {
		memcpy(vh->id3.header+vh->id3.curpos, vh->inbuf, vh->inbuflen);
		vh->id3.curpos += vh->inbuflen;
		return -1;
	}
	memcpy(vh->id3.header+vh->id3.curpos, vh->inbuf, (10 - vh->id3.curpos));
	vh->id3.curpos += (10 - vh->id3.curpos);

	if(strncmp(vh->id3.header, "ID3", 3) == 0) {
		vh->id3.flags = vh->id3.header[5];
		vh->id3.size = vh->id3.header[9] + (vh->id3.header[8]<<7) + (vh->id3.header[7]<<14) + (vh->id3.header[6]<<21);
		vh->id3.hasfooter = (vh->id3.flags>>4)&0x1;
		return (vh->id3.size + 10 + (10 * vh->id3.hasfooter));
	} else {
		return 0;
	}
}
